#ifndef _IRI_CAN_H
#define _IRI_CAN_H

#include "comm.h"
#include "mutex.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <net/if.h>
#include <linux/can.h>
#include <linux/can/bcm.h>
#include <linux/can/raw.h>

#include "threadserver.h"
#include "eventserver.h"
#include <queue>

/**
 * \brief CAN driver
 *
 */

class CCAN : protected CComm
{
  private:
    // can device identifier
    int can_socket_fd;
    unsigned int req_can_id;
    // thread attributes
    CThreadServer *thread_server;
    std::string can_thread_id;
    // event attributes
    CEventServer *event_server;
    std::string new_frame_event_id;// public event
    std::string frame_error_event_id;// public event
    std::string data_requested_event_id;// public event
    std::string finish_can_thread_event_id;// private event
    // internal received frame buffer
    std::queue<struct can_frame> rx_frames;
    // reception can_id filters
    struct can_filter *rx_filters;
    int num_filters;
    CMutex can_access;
  protected:
     /**
      * \brief Function to actually open the device
      *
      * This function is called automatically when the base class open() function
      * is called. By default, it initializes the socket file descriptor using the
      * socket() system call. But, if the object has been created by the 
      * create_socket() function, this function does nothing.
      *
      * This class does not need any parameter since the socket is created by 
      * default using a TCP/IP protocol and the AF_INET socket family. So the
      * argument passed to this function is ignored and may be NULL.
      * 
      * This function can throw any CSocketException object exception or the generic 
      * CCommException class.
      *
      * \param comm_dev this parameter is ignored in this case and can be set to NULL.
      *                 It is only keeped for backward compatiblity with the CComm
      *                 class.
      */
    virtual void hard_open(void *comm_dev=NULL);
    /**
      * \brief Function to actually configure the device
      * 
      * This function is called automatically when the base class config() function
      * is called. In this case this function does nothing, since all the necessay
      * parameters are provide to the open() function. However, it is necessary to 
      * call the config() function to successfully change the internal state of the
      * communication device.
      *
      * The provided parameter in this case can be NULL since no configuration is
      * needed. This function can throw any CCommException object exception or else 
      * any exception class defined by the inherited class.
      *
      * \param config This parameter is not used since no configuration is required.
      *               It is only keeped for backward compatiblity with the CComm
      *               class.
      */
     virtual void hard_config(void *config=NULL);
    /**
     * \brief Function to actually read from the device
     *
     * This function is automatically called when the new data received event is 
     * activated. The read() function from the base class gets data from the 
     * internal queue, so this function is not used. It must try to read the 
     * ammount of data specified and store it in the data buffer provided without 
     * blocking. Also, it must return the number of bytes actually read from the 
     * devicve, since they may be different than the desired value.
     *
     * In case of any error, this function throws a CSocketException exception.
     *
     * \param data a reference to the buffer where the received data must be 
     * copied. The necessary memory for this buffer must be allocated before 
     * calling this function and have enough size to store all the desired data.
     * If this buffer is not initialized, the function throws an exception.
     *
     * \param len a positive interger that indicates the number of byte to be
     * read from the communication device. This value must be at most the length 
     * of the data buffer provided to the function.
     *
     * \return an integer with the number of bytes actually read. These number 
     * coincide with the desired number if there is enough data in the internal 
     * queue, but it could be smaller if not.
     */
    virtual int hard_read(unsigned char *data, int len);
    /**
     * \brief Hard write function
     *
     * This function is automatically called when the base class write() function
     * is called. It must try to write the desired ammount of data to the communication 
     * device without blocking. Also it must return the number of bytes actually
     * written to the communication device since they may be different from the
     * desired value.
     *
     * In case of any error, this function throws a CSocketException exception.
     *
     * \param data a reference to the buffer with the data must be send to the 
     * device. The necessary memory for this buffer must be allocated before 
     * calling this function and have enough size to store all the desired data.
     * If this buffer is not initialized, the function throws an exception.
     *
     * \param len a positive interger that indicates the number of byte to be
     * written to the communication device. This value must be at most the length 
     * of the data buffer provided to the function.
     *
     * \return an integer with the number of bytes actually written. These number 
     * coincide with the desired number if there is enough data in the internal 
     * queue, but it could be smaller if not.
     */
    virtual int hard_write(unsigned char *data, int len);
    /**
     * \brief Function to actually get the number of bytes availables
     *
     * This function is called when the new data received event is activated.
     * It must get the number of data bytes available from the communication
     * device and return its value without blocking.
     *
     * In case of any error, this function throws a CSocketException exception.
     *
     * \return an integer with the number of bytes available in the receive queue. 
     * This value can be 0 if there is no data available, but there is ni upper
     * limit in its value.
     */
    virtual int hard_get_num_data(void);
    /**
     * \brief Function to actually wait for a given communication event 
     *
     * This function is called in the internal communciation thread to wait for 
     * any event on the communuication device. It must check for any event on the
     * device (reception, end of transmission or error) and return the
     * corresponding identifier. When an event is activated, this function must 
     * return to allow the base class to handle it, and the it is called again
     * to wait for the next event.
     *
     * This function can throw any CCommException object exception or else any
     * exception class defined by the inherited class.
     *
     * \return -1 if there has been any error or else the identifier of the 
     *         communciation event:
     *
     * - 1 for the new data received event
     * - 2 for the error event
     */
    virtual int hard_wait_comm_event(void);
    /** 
     * \brief Hard close function
     *
     * This function is called when the base class close() funciton is called. It
     * must free the device handle initialized by the open() function and also free
     * any other resource allocated by the inherited class.
     *
     * This function can throw any CCommException object exception or else any
     * exception class defined by the inherited class.
     */
    virtual void hard_close(void);
    /** 
     * \brief
     */
    static void *can_thread(void *param);
  public:
    /**
     * \brief default constructor
     *
     * This constructor creates a new CSocket object. It does not open a physical 
     * socket, it only allocates the necessary resources to use it.
     *
     * \param comm_id  A null terminated string which identifies the 
     *                 communications device. This string is used to created a
     *                 unique identifier for all the threads and events of the
     *                 class.
     */
    CCAN(const std::string &comm_id);
    /** 
     * \brief
     */
    void open(const std::string &can_dev);
    /** 
     * \brief
     */
    std::string get_new_frame_event_id(void);
    /** 
     * \brief
     */
    std::string get_frame_error_event_id(void);
    /** 
     * \brief
     */
    std::string get_data_requested_event_id(void);
    /** 
     * \brief
     */
    unsigned int get_requested_can_id(void);
    /** 
     * \brief
     */
    unsigned long get_last_rx_timestamp(void);
    /** 
     * \brief
     */
    void write_frame(unsigned int can_id, unsigned char *data, int len);
    /** 
     * \brief
     */
    void read_frame(unsigned int *can_id, unsigned char *data, int *len);
    /** 
     * \brief
     */
    void request_frame(unsigned int can_id, int len);
    /** 
     * \brief
     */
    void add_id_filter(unsigned short int can_id,bool invert);
    /** 
     * \brief
     */
    void add_id_filter(unsigned short int can_id, unsigned short int mask,bool invert);
    /** 
     * \brief
     */
    void clear_id_filters(void); 
    /** 
     * \brief
     */
    void close(void);
    /**
     * \brief Destructor
     *
     * This destructor does nothing. The base class destructor is the one in 
     * charge of freeing all the allocated resources.
     */
    virtual ~CCAN();
};

#endif
